import hashlib
import traceback
from http import HTTPStatus

import aiohttp
import time

from common import Message, PushResult
from exceptions import AuthSignFailed


class IGeTui:
    SIGN_URL = 'https://restapi.getui.com/v1/{app_id}/auth_sign'
    PUSH_SINGLE_URL = 'https://restapi.getui.com/v1/{app_id}/push_single'

    def __init__(self, app_id, app_key, master_secret, loop=None):
        self.app_id = app_id
        self.app_key = app_key
        self.master_secret = master_secret

        self.loop = loop
        self.session = None

        # sign
        self.sign_url = self.SIGN_URL.format(app_id=self.app_id)
        self.sign_timestamp = None
        self.auth_token = None

        # push
        self.push_single_url = self.PUSH_SINGLE_URL.format(app_id=self.app_id)

    async def auth_sign(self):
        """用户身份验证通过获得auth_token权限令牌，后面的请求都需要带上auth_token
        """
        if self.session is None:
            self.session = aiohttp.ClientSession(loop=self.loop)

        self.sign_timestamp = int(time.time() * 1000)
        raw_sign = self.app_key + str(self.sign_timestamp) + self.master_secret
        sign = hashlib.sha256(raw_sign.encode()).hexdigest()

        sign_params = {
            'appkey': self.app_key,
            'timestamp': self.sign_timestamp,
            'sign': sign,
        }
        json_result = await self._post(self.sign_url, json=sign_params)

        if 'auth_token' not in json_result:
            raise AuthSignFailed(f'Reason: {json_result.get("result")}')
        self.auth_token = json_result['auth_token']

    async def push(self, message: Message) -> PushResult:
        """推送消息
        """
        assert self.auth_token is not None, \
            'No auth_token, cannot send request'

        try:
            json_result = await self._post(self.push_single_url,
                                           json=message.to_params(self.app_key))
        except aiohttp.ClientError:
            return PushResult(PushResult.HTTP_REQUEST_FAILED,
                              description=traceback.format_exc())
        return PushResult(json_result.get('result'),
                          json_result.get('desc'),
                          json_result.get('taskid'),
                          json_result.get('status'))

    async def _post(self, url, json=None):
        headers = dict()
        if self.auth_token:
            headers.update({'authtoken': self.auth_token})
        async with self.session.post(
                url, json=json, headers=headers) as response:
            assert response.status == HTTPStatus.OK
            return await response.json(content_type='text/html')

    async def close(self):
        if self.session:
            await self.session.close()



class Template:
    """消息模板
    有四种:
    1. notification, 点击通知直接打开应用
    2. link, 用户点击时, 可以打开指定的网页
    3. notypopload, 弹窗通知, 用户点击可直接下载文件/应用, 可用于检测版本更新
    4. transmission, 透传消, 是指消息传递到客户端只有消息内容, 展现的形式由客户端自行定义
    """
    type = None

    def to_dict(self):
        raise NotImplementedError


class NotificationTemplate(Template):
    type = 'notification'

    def __init__(self,
                 style: dict,
                 transmission_type: bool=None,
                 transmission_content: str=None,
                 duration_begin: str=None,
                 duration_end: str=None):
        """
        除了 style, 都可以缺省
        :param style: {
                   "type": 0,
                   "text": "请填写通知内容",
                   "title": "请填写通知标题",
                   "logo": "logo.png",
                   "logourl": "http://xxxx/a.png",
                   "is_ring": true,
                   "is_vibrate": true,
                   "is_clearable": true
                 }
        :param transmission_type: 收到消息是否立即启动应用, true 为立即启动,
                                  false 则广播等待启动
        :param transmission_content: 透传内容, 由用户自定义及客户端自行处理
        :param duration_begin: 设定展示开始时间, 格式为 yyyy-MM-dd HH:mm:ss
        :param duration_end: 展示结束时间，格式为 yyyy-MM-dd HH:mm:ss
        """
        self.style = style
        if 'type' not in self.style:
            self.style['type'] = 0
        self.transmission_type = transmission_type
        self.transmission_content = transmission_content
        self.duration_begin = duration_begin
        self.duration_end = duration_end

    def to_dict(self):
        result = dict()
        result['style'] = self.style
        if self.transmission_type is not None:
            result['transmission_type'] = self.transmission_type
        if self.transmission_content is not None:
            result['transmission_content'] = self.transmission_content
        if self.duration_begin is not None:
            result['duration_begin'] = self.duration_begin
        if self.duration_end is not None:
            result['duration_end'] = self.duration_end
        return result



import uuid
from template import Template


class Message:

    def __init__(self,
                 client_id: str,
                 template: Template,
                 is_offline: bool = False,
                 offline_expire_time: int = 0,
                 message_id: str = None,
                 push_info: dict = None):
        """
        :param client_id: 客户端的 id
        :param template: 消息内容模板
        :param is_offline: 是否离线发送
        :param offline_expire_time: 离线消息的有效期(单位: ms)
        :param message_id: 长度为 10~32
        """
        self.client_id = client_id
        self.template = template
        self.is_offline = is_offline
        self.offline_expire_time = offline_expire_time

        if message_id is None:
            self.message_id = uuid.uuid4().hex  # 长度限制为 32
        else:
            self.message_id = message_id

        # 当手机为 ios，并且为离线的时候, 可是通过 push_info 发送消息
        self.push_info = push_info

        self.duration = 0

    def to_params(self, app_key: str):
        raise NotImplementedError


class ToSingleMessage(Message):
    """单独推送的消息
    """

    def to_params(self, app_key: str):
        params = {
            'cid': self.client_id,
            'requestid': self.message_id,
            'message': {
                'appkey': app_key,
                'is_offline': self.is_offline,
                'offline_expire_time': self.offline_expire_time,
                'msgtype': self.template.type,
            },
            self.template.type: self.template.to_dict(),
        }
        if self.push_info:
            params['push_info'] = self.push_info
        return params


class PushResult:
    OK = 'ok'
    INVALID_CLIENT_ID = 'no_user'
    HTTP_REQUEST_FAILED = 'http_request_failed'

    def __init__(self,
                 result: str,
                 description: str=None,
                 task_id: str=None,
                 status: str=None):
        self.result = result
        self.description = description
        self.task_id = task_id
        self.status = status

    def __str__(self):
        return f'PushResult<' \
               f'result: {self.result}, ' \
               f'description: {self.description}, ' \
               f'task_id: {self.task_id}, ' \
               f'status: {self.status}, ' \
               f'>'

    @property
    def is_successful(self):
        return self.status == self.OK

    @property
    def is_client_id_invalid(self):
        return self.status == self.INVALID_CLIENT_ID



from aiogetui.client import IGeTui
from aiogetui.common import ToSingleMessage
from aiogetui.template import NotificationTemplate

__all__ = [
    'IGeTui',
    'ToSingleMessage',
    'NotificationTemplate',
]
